<?php
if (!defined('MEDIAWIKI')) die("MediaWiki extensions cannot be run directly.");
if (!defined('WikiDB')) {
	die("The WikiDB extension has not been set up correctly.<br>'"
		. basename(__FILE__) . "' should not be included directly.<br>Instead, "
		. "add the following to LocalSettings.php: "
		. "<tt><b>include_once(\"WikiDB/WikiDB.php\");</b></tt>");
}
/**
 * Functions to handle the new parser tags implemented by WikiDB.
 *
 * IMPLEMENTATION NOTE:
 *	These tags use $Parser = NULL as the third arg, instead of &$Parser.
 *	This is because the $Parser argument was not available in MW 1.5.6 and earlier.
 *  I'm not sure when it was introduced, but I'm going to keep it like this for the
 *	time being to ensure backwards compatability.
 *	If the argument is not passed to the funtion, we pull it from the $GLOBALS array
 *	instead.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2013, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: 792 $
 */

WikiDB_RegisterRevision('$Rev: 792 $');

// wfWikiDB_GuessTypes() - Handler for the <guesstypes> tag.
//
// Purpose:
// 	This tag is mainly for testing purposes.  It allows you to provide wiki-based
//  input and see which type WikiDB will treat it as, if the field is otherwise
//  untyped.  Primarily for use by people creating their own data types for use
//  within WikiDB.
//
// Arguments:
//	None.
//
// Tag contents:
// 	A valid line is any line with a colon.  The text before the colon is a
//  description which is displayed in the first column of the resulting table
//  (useful when there are multiple tests within the same tag) and the text after it
//  is the string to be parsed.  Multiple colons are allowed - the two components
//  are split on the first colon, and subsequent occurences of a colon are part of
//  the text to be parsed (so the description may not contain a colon.
//  Non-valid lines are ignored, and not output at all.
//
// Output:
//	A table with one row per valid line, with the following columns: Description,
//  Original value unrendered, original value rendered, type, typed value
//  unrendered, typed value rendered.
//	* Description - as entered on the page.
//	* Original value - as entered on the page.
//	* Type - the name of the type, as guessed by WikiDB.
//	* Typed value - the result of formatting the original value as if it was the
//    guessed type.
//  For the 'Original Value' and 'Typed Value', there are two columns output:
//	* Unrendered - the literal wikitext, unparsed.
//	* Rendered - the rendered output, as it would appear in the page.
//
// TODO: Use wfMsg() for text.
// TODO: Use CSS classes rather than in-line styling.
	function wfWikiDB_GuessTypes($Input, $Args, $Parser = NULL) {
		global $wgOut;
		global $wgpWikiDBParsingForSave;

	// If the page is being parsed for a save operation, then do nothing.
		if ($wgpWikiDBParsingForSave)
			return "";

	// Get $Parser if it was not supplied.
	    if (!$Parser)
			$Parser =& $GLOBALS['wgParser'];

	// Explode the input into an array of separate lines.
		$Lines = explode("\n", $Input);

	// Create an array of formatted lines.  This contains the data that we need to
	// output for each line that contained a valid description/input-text pair, i.e.
	// any line containing a colon.
		$FormattedLines = array();
		foreach ($Lines as $Line) {
		// If line starts with = then it is the correct answer for the previous
		// line (useful for unit-testing).  In which case, we add it to the
		// output of the previous line in the $FormattedLines array.
		// If line does not contain a colon then it the whole string is the expected
		// output.
		// If it contains a colon, everything before the first colon is the
		// expected type and everything after the first colon is the expected output.
		// The expected type may be left blank, which is the same as omitting it
		// completely, except it allows you to supply an expected output that
		// contains a semi-colon.
		// TODO: Is /s switch needed?
			if (preg_match('/^=(.*)(?::(.*))?$/Us', $Line, $Matches)) {
				$Count = count($FormattedLines) - 1;
			// Only check the item if there is a previous line that this can apply
			// to - otherwise it is ignored.
				if ($Count >= 0) {
				// If there was no colon, the output is in the first element and the
				// second element is blank.  We therefore bump things so that the
				// output is in the second element and the first element (the type)
				// is blank.
					if (!isset($Matches[2])) {
						$Matches[2] = $Matches[1];
						$Matches[1] = "";
					}

				// Normalise the type name.  This ensures it matches the rules for
				// naming types (currently case-insensitive) and will also trim
				// whitespace (therefore treating it as if nothing was entered).
					$Matches[1] = WikiDB_TypeHandler::NormaliseTypeName($Matches[1]);

				// If a type was filled in, then store it as the expected type.
				// We don't care whether the type is defined, just whether it matches
				// the guessed type or not.
					if ($Matches[1] != "")
						$FormattedLines[$Count]['ExpectedType'] = $Matches[1];

					$FormattedLines[$Count]['ExpectedOutput'] = $Matches[2];
				}
			}
		// Otherwise, line is a new data definition line, which should have its type
		// guessed, in which case tidy it and add it to the $FormattedLines array.
		// TODO: Is /s switch needed?
			elseif (preg_match('/^(.*):(.*)$/Us', $Line, $Matches)) {
				$Description = trim($Matches[1]);
				$Data = trim($Matches[2]);
				$Type = WikiDB_TypeHandler::GuessType($Data);
				$FormattedData =
						WikiDB_TypeHandler::FormatTypeForDisplay($Type, $Data);
				$SortFormat =
						WikiDB_TypeHandler::FormatTypeForSorting($Type, $Data);

				$FormattedLines[] = array(
											'Description' 	=> $Description,
											'Data'			=> $Data,
											'Type'			=> $Type,
											'FormattedData'	=> $FormattedData,
											'SortFormat'	=> $SortFormat,
										);
			}
		}
		$Output = "";

	// If the input contained any valid lines, then output the table.
		if (count($FormattedLines) > 0) {
		// Open table tag:
			$Output .= '<table style="border-collapse: collapse;" border="1" '
					 . 'cellpadding="5">' . "\n";

		// Header rows:
			$Output .= "\t" . '<tr bgcolor="#EEEEEE">'
					 . '<th rowspan="2">Description</th>'
					 . '<th colspan="2">Original value</th>'
					 . '<th rowspan="2">Type</th>'
					 . '<th colspan="2">Typed value</th>'
					 . '<th rowspan="2">Formatted for sorting<br><small>Spaces are currently collapsed, for ease of reading. Check source for full string.</small></th>'
					 . '</tr>' . "\n";
			$Output .= "\t"
					 . '<tr bgcolor="#EEEEEE">'
					 . '<th>Unrendered</th>'
					 . '<th>Rendered</th>'
					 . '<th>Unrendered</th>'
					 . '<th>Rendered</th>'
					 . '</tr>' . "\n";

		// One row per valid line:
			foreach ($FormattedLines as $Line) {
				$Output .= "\t<tr>\n";

			// Description of this item
				$Output .= "\t\t<td>"
						 . WikiDB_Parse($Line['Description'], $Parser, false)
						 . "</td>\n";
			// Original value: unrendered
				$Output .= "\t\t<td>"
						 . htmlspecialchars($Line['Data'])
						 . "</td>\n";
			// Original value: rendered
				$Output .= "\t\t<td>"
						 . WikiDB_Parse($Line['Data'], $Parser, false)
						 . "</td>\n";

			// Guessed type (highlighted in green or red to indicate whether the
			// result was correct or not, if an expected type was given.
				if (isset($Line['ExpectedType']))
					$Style = pStyleCorrectness($Line['ExpectedType'], $Line['Type']);
				else
					$Style = "";
				$Output .= "\t\t<td" . $Style . ">" . $Line['Type'] . "</td>\n";
			// Typed value: unrendered
			// Highlighted in green or red to indicate whether the
			// result was correct or not, if an expected output was given.
				if (isset($Line['ExpectedOutput'])) {
					$Style = pStyleCorrectness($Line['ExpectedOutput'],
											   $Line['FormattedData']);
				}
				else
					$Style = "";
				$Output .= "\t\t<td" . $Style . ">"
						 . htmlspecialchars($Line['FormattedData'])
						 . "</td>\n";
			// Typed value: rendered
				$Output .= "\t\t<td>"
						 . WikiDB_Parse($Line['FormattedData'], $Parser, false)
						 . "</td>\n";

			// Stored value:
				$Output .= "\t\t<td>"
						 . WikiDB_Parse("'''['''<code><nowiki>" . $Line['SortFormat'] 
										. "</nowiki></code>''']'''", $Parser, false)
						 . "</td>\n";
				$Output .= "\t</tr>\n";
			}

		// Close table tag:
			$Output .= "</table>\n";
		}

		return $Output;
	}

// pStyleCorrectness()
// Returns a style string depending on whether the two items are equal or not.
	function pStyleCorrectness($Val1, $Val2) {
//		var_dump($Val1); print(" - "); var_dump($Val2); print("\n");
		$Style = ' style="background-color: ';
		if ($Val1 == $Val2)
			$Style .= "#D0FFD0";
		else
			$Style .= "#FFD0D0";
		$Style .= ';"';

		return $Style;
	}

// wfWikiDB_DataTag() - Handler for the <data> tag.
//
// Purpose:
//	To define rows of data belonging to the specified table.
//
// Arguments:
//	table:
//		The name of the table the data should be added to. Required.
//	template:
//		The name of a template that the data should be passed to for display.
//		Additionally, "none" (the default value) may be used to supress all output,
//		and "default" may be used to display the data in the default WikiDB style.
//
// Tag contents:
//	See WikiDB_Parser::ParseDataTag().
//
// Output:
//	Nothing if no template argument is supplied, or template="none".
//	A table showing the data in the standard WikiDB format if template="default".
//	The output generated by the specified template (with the fields passed as
//  template parameters) if neither of the above.
//
// TODO: Use wfMsg() for text.
	function wfWikiDB_DataTag($Input, $Args, $Parser = NULL) {
		global $wgOut, $wgTitle;
		global $wgpWikiDBParsingForSave;

	// This tag has no effect in the MediaWiki namespace, so return the contents
	// unmodified.  $Input just contains the tag contents, so we need to 
	// re-wrap it in the <data> tag, including the supplied arguments, in order to
	// undo the effects of the tag.
	// Note that I haven't used htmlspecialchars() here, because I believe this
	// will be done automatically by MW before the result is output, and it shouldn't
	// happen twice.
		if ($wgTitle->getNamespace() == NS_MEDIAWIKI) {
			$Output = "<data";
			foreach ($Args as $ArgName => $ArgValue)
				$Output .= ' ' . $ArgName . '="' . $ArgValue . '"';
			$Output .= ">" . $Input . "</data>";
			return $Output;
		}

	// Get $Parser if it was not supplied.
	    if (!$Parser)
			$Parser =& $GLOBALS['wgParser'];

	// Normalise the tag arguments so that all variables are set
	// and appropriate defaults are filled in where necessary.
		$Args = pWikiDB_DataTag_NormaliseArgs($Args);

	// If no table argument, then the data is unusable, so we print a warning and
	// exit.
		if ($Args['table'] == "") {
			return WikiDB_Parse("''No table specified in &lt;data&gt; tag!''",
								$Parser);
		}

	// If the page is being parsed for a save operation, call the function
	// that does the DB update and exit.
		if ($wgpWikiDBParsingForSave) {
			WikiDB::pAddDataDefinition($wgTitle, $Input, $Args);
			return "";
		}

	// If the template is "none", then the data is not displayed, so we return an
	// empty string.
		if ($Args['template'] == "none")
			return "";

	// Parse the tag contents using the standard WikiDB data parsing algorithm.
		$Table = new WikiDB_Table($Args['table']);
		$Data = WikiDB_Parser::ParseDataTag(
											$Input,
											$Args['fields'],
											$Args['separator']
										);

	// Add meta-data fields.
		foreach ($Data as $Key => $Row) {
			$Data[$Key] = array('Data' => $Row);
			$Data[$Key]['MetaData']['_SourceArticle'] = $wgTitle->getFullText();
			$Data[$Key]['MetaData']['_Row'] = $Key + 1;
		}

	// If using the 'default' template, output the data in the standard way.
		if ($Args['template'] == "default")
			$Output = $Table->FormatTableData($Data);

	// Otherwise, call the appropriate template with each field as a named parameter.
		else {
			$Output = "";

		// If a header template is specified, include that.
			if ($Args['thead'] != "")
				$Output .= "{{" . $Args['thead'] . "}}";

		// Include the body, once for each row of data.
			foreach ($Data as $RowData) {
				$Output .= "{{" . $Args['template'];
				if (count($RowData) > 0) {
					foreach ($RowData['MetaData'] as $Name => $Value)
						$Output .= "|" . $Name . "=" . $Value;
					foreach ($RowData['Data'] as $Name => $Value) {
						$Output .= "|" . $Name . "="
								 . $Table->FormatFieldValue($Name, $Value);
					}
				}
				$Output .= "}}";
			}

		// If a footer template is specified, include that.
			if ($Args['tfoot'] != "")
				$Output .= "{{" . $Args['tfoot'] . "}}";
		}

		return WikiDB_Parse($Output, $Parser);
	}


// wfWikiDB_RepeatTag() - Handler for the <repeat> tag.
//
// Purpose:
//	To output data from a table in the specified format.
//	THIS TAG IS CURRENTLY VERY EXPERIMENTAL, AND LIKELY TO CHANGE DRASTICALLY!
//
// Arguments:
//	table:
//		The name of the table to pull the data from.
//	sort:
//		Sort criteria (optional).  Uses default (undefined) sort order if omitted.
//	criteria:
//		Filter criteria (optional).  Defaults to all records if omitted.
//
// Tag contents:
//	May be left blank.  If not then the contents are repeated for each row of the
//	resulting data-set, with variables expanded in a similar way to template
//  parameters.
//
// Output:
//	If the tag has no contents then the resulting data-set is output in the standard
//  WikiDB format.  Otherwise the contents are repeated for each row of data, with
//  variables expanded.
//
// TODO: Use wfMsg() for text.
	function wfWikiDB_RepeatTag($Input, $Args, $Parser = NULL) {
		global $wgOut;
		global $wgpWikiDBParsingForSave;

	// If the page is being parsed for a save operation, then do nothing.
		if ($wgpWikiDBParsingForSave)
			return "";

	// Get $Parser if it was not supplied.
	    if (!$Parser)
			$Parser =& $GLOBALS['wgParser'];

	// Repeat tags change all the time, so disable caching.
		$Parser->disableCache();


	// Get table argument (required)
		if (isset($Args['table']))
			$TableName = $Args['table'];
		else
			$TableName = "";

	// If no table argument, then there is no way of retrieving the data, so we
	// print a warning and exit.
		if ($TableName == "")
			return WikiDB_Parse("''No table specified in &lt;repeat&gt; tag!''",
								$Parser);


	// Get sort argument (optional - default to unsorted)
		if (isset($Args['sort']))
			$Sort = $Args['sort'];
		else
			$Sort = "";


	// Get criteria argument (optional - default to unfiltered)
		if (isset($Args['criteria']))
			$Where = $Args['criteria'];
		else
			$Where = "";


	// Perform the query and get the resulting data set.
		$Query = new WikiDB_Query($TableName, $Where, $Sort);
		$Data = $Query->Run();

	// If returned data is not an array, then an error occurred, so we output the
	// error.
		if (!is_array($Data))
			return $Data;

	// TODO: Should not be needed here.  Rework later code so it doesn't need a
	// 		 table object, as the query may be run on several tables (JOIN).
		$Table = new WikiDB_Table($TableName);

	// If tag contents are empty, display in the standard format.
	// TODO: FormatTableData should be a static function!
		if (trim($Input) == "")
			$Output = $Table->FormatTableData($Data);

	// If tag contents are non-empty then use it as a template for each row
	// of the returned data, expanding variables where necessary.
		else {
			$arrInput = WikiDB_Parser::ParseQueryTag($Input);
			$Data = $Table->NormaliseData($Data);

			$Output = "";
			foreach ($Data as $Row)
				$Output .= pWikiDB_ExpandVariables($arrInput['body'], $Row);

		// Add any header/footer that was specified.
			$Output = $arrInput['header'] . $Output . $arrInput['footer'];
		}

	// Return the resulting output, parsed as wiki text.
		return WikiDB_Parse($Output, $Parser);
	}

// pWikiDB_DataTag_NormaliseArgs()
// Takes an array containing the arguments which may be supplied to a <data> tag
// and normalises them to ensure that (a) all args are present (b) sensible defaults
// are used and (c) any arguments which may have more than one representation are
// normalised to a standard form (e.g. all case-insensitive args become lower-case).
	function pWikiDB_DataTag_NormaliseArgs($Args) {
	// If the argument list is not an array, then it is completely invalid, so
	// we replace it with an empty array which will allow the defaults to be
	// filled in correctly by the subsequent code.
		if (!is_array($Args))
			$Args = array();

	//////////////////////////////////////////////////
	// 'table' - the table that the data belongs in.
	// Required, default = "" (no table specfied)

		if (!isset($Args['table']))
			$Args['table'] = "";
		else
			$Args['table'] = trim($Args['table']);

	//////////////////////////////////////////////////
	// 'thead'/'tfoot' - templates to use just once before/after the data rows.
	// If template is one of the built-in defaults, or 'fields' is not
	// set, then this will be ignored (we fill it in here, but reset it to "" if any
	// of these conditions are met, which is why these arguments are checked first).

		if (!isset($Args['thead']))
			$Args['thead'] = "";
		else
			$Args['thead'] = trim($Args['thead']);

		if (!isset($Args['tfoot']))
			$Args['tfoot'] = "";
		else
			$Args['tfoot'] = trim($Args['tfoot']);

	//////////////////////////////////////////////////
	// 'template' - the template used to display the data.
	// Possible values are 'none' (not displayed), 'default' (displayed in the
	// standard built-in format) or else the name of a template to use.  This
	// template will be passed all the field values by name, and will be included
	// once for each row of data within the tag.
	// Optional, default = "none"

	// If not present, set to default value.
		if (!isset($Args['template']))
			$Args['template'] = "none";
	// Otherwise, trim any leading/trailing whitespace from the name
		else
			$Args['template'] = trim($Args['template']);

	// Normalise the two special internal values, which are
	// case-insensitive.  If either of these two values are used, then we also
	// remove any thead/tfoot settings as these should not be used in those
	// instances.
		switch (strtolower($Args['template'])) {
			case "none":
			case "default":
				$Args['template'] = strtolower($Args['template']);
				$Args['thead'] = "";
				$Args['tfoot'] = "";
				break;
		}

	//////////////////////////////////////////////////
	// 'fields' - A list of field names indicating the columns being defined by this
	// data tag.  If this argument is present then we are in multi-row mode, meaning
	// that each row of the <data> tag is a row of data, with fields separated by
	// the specified separator string (or "," if none specified).  The fields in each
	// row contain data for the fields named in this list.  If a row has too many
	// fields, later ones are discarded.  If it has too few they are set to "" (empty
	// string).
	// In single-row mode, no 'fields' argument is used, and instead each row is
	// a "name=value" pair.

	// If the 'fields' argument is not present, set it to NULL. We use NULL rather
	// than "" because we need to distinguish between 'not present' and 'present but
	// empty'.  The former means we are in single-row mode, the latter means we are
	// in multi-row mode but no fields were specified.
		if (!isset($Args['fields'])) {
			$Args['fields'] = NULL;
		// In single-row mode, there is no need for header/footer templates as there
		// is only one row.
		// TODO: Actually, it might be useful to allow this, even though it's not
		// 		 needed, for example so that you can use the same template for the
		//		 body in multi/single-row modes and still wrap it in the appropriate
		// 		 header/footer - though note that this can still be acheived using
		// 		 standard in-line template tags before/after the <data> tag, which
		//		 might be better-practice.
			$Args['thead'] = "";
			$Args['tfoot'] = "";
		}

	//////////////////////////////////////////////////
	// 'separator' - the character string used to separate field values when
	// in multi-row mode.  This is only relevant if 'fields' is filled in, but
	// we make sure it is filled in and valid in all cases.
	// Optional, default = ","

	// If not present, or blank, set to default value.  Note that we don't
	// trim() the setting, as space could be a valid separator.
	// TODO: Newline cannot be used, so perhaps we should remove them.  That
	// 		 said, it may make more sense not to 'fix' the string in this way, as the
	// 		 fix may create unwanted behaviour and it is unlikely that a newline
	//		 will be inserted accidentally inside a quoted attribute.
		if (!isset($Args['separator']) || $Args['separator'] == "")
			$Args['separator'] = ",";

	//////////////////////////////////////////////////

	// Return the normalised set of arguments.
		return $Args;
	}