<?php
if (!defined('MEDIAWIKI')) die("MediaWiki extensions cannot be run directly.");
if (!defined('WikiDB')) {
	die("The WikiDB extension has not been set up correctly.<br>'"
		. basename(__FILE__) . "' should not be included directly.<br>Instead, "
		. "add the following to LocalSettings.php: "
		. "<tt><b>include_once(\"WikiDB/WikiDB.php\");</b></tt>");
}
/**
 * WikiDB field type handler - for managing and accessing data types for WikiDB
 * table fields.  Types are extensible - it is easy to add new ones via
 * this class.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2013, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: $
 */

WikiDB_RegisterRevision('$Rev: $');

class WikiDB_TypeHandler {

// pManageTypes()
//
// Private function to handle data type validation & formatting.
	function pManageTypes($Action, $TypeName = NULL, $CallbackFunction = NULL) {
	// Create built-in types.
		static $Types = array(
			'wikistring'	=> array("pWikiDB_BuiltInTypeHandlers",
									 "pWikiStringTypeHandler"),
			'string' 		=> array("pWikiDB_BuiltInTypeHandlers",
									 "pStringTypeHandler"),
			'integer' 		=> array("pWikiDB_BuiltInTypeHandlers",
									 "pIntegerTypeHandler"),
			'number' 		=> array("pWikiDB_BuiltInTypeHandlers",
									 "pNumberTypeHandler"),
			'int' 			=> array("pWikiDB_BuiltInTypeHandlers",
									 "pIntegerTypeHandler"), // Alias for 'integer'.
//			'date' 			=> array("pWikiDB_BuiltInTypeHandlers",
//									 "pStringTypeHandler"),
			'image' 		=> array("pWikiDB_BuiltInTypeHandlers",
									 "pImageTypeHandler"),
			'link' 			=> array("pWikiDB_BuiltInTypeHandlers",
									 "pLinkTypeHandler"),
		);

		$TypeName = WikiDB_TypeHandler::NormaliseTypeName($TypeName);

		switch ($Action) {
			case pWIKIDB_AddType:
				$Types[$TypeName] = $CallbackFunction;
				break;
			case pWIKIDB_GetCallbackFunction:
				if (isset($Types[$TypeName]))
					return $Types[$TypeName];
				else
					return NULL;
				break;
			case pWIKIDB_GetAllTypes:
				return $Types;
				break;
			case pWIKIDB_TypeDefined:
				if (isset($Types[$TypeName]))
					return true;
				else
					return false;
				break;
		}
	}

	/* static */
	function NormaliseTypeName($TypeName) {
		return strtolower(trim($TypeName));
	}

	/* static */
	function AddType($TypeName, $CallbackFunction) {
		return WikiDB_TypeHandler::pManageTypes(pWIKIDB_AddType, $TypeName,
										  $CallbackFunction);
	}

	/* static */
	function TypeDefined($TypeName) {
		return WikiDB_TypeHandler::pManageTypes(pWIKIDB_TypeDefined, $TypeName);
	}

	/* static */
	function IsOfType($TypeName, $Value) {
		$CallbackFunction = WikiDB_TypeHandler::pManageTypes(
													pWIKIDB_GetCallbackFunction,
													$TypeName
												);

		if (is_callable($CallbackFunction)) {
			return call_user_func($CallbackFunction, WIKIDB_Validate,
								  $Value, array());
		}
	// If the type doesn't exist, then the field is valid.
		else
			return true;
	}

	/* static */
	function FormatTypeForDisplay($TypeName, $Value, $Options = array()) {
		$CallbackFunction = WikiDB_TypeHandler::pManageTypes(
														pWIKIDB_GetCallbackFunction,
														$TypeName
													);

	// Always trim() the supplied value to remove whitespace.
		$Value = trim($Value);

		if (is_callable($CallbackFunction)) {
			if (call_user_func($CallbackFunction, WIKIDB_Validate,
							   $Value, array()))
			{
				return call_user_func($CallbackFunction, WIKIDB_FormatForDisplay,
									  $Value, $Options);
			}
			else
				return "??";
		}
		else
			return $Value;
	}

	/* static */
	function FormatTypeForSorting($TypeName, $Value) {
		$CallbackFunction = WikiDB_TypeHandler::pManageTypes(
														pWIKIDB_GetCallbackFunction,
													    $TypeName
													);

		if (is_callable($CallbackFunction)) {
			return call_user_func($CallbackFunction, WIKIDB_FormatForSorting,
								  $Value, array());
		}
		else
			return $Value;
	}

// GuessType()
//
// Guesses the field type of the supplied field data.
// It does this by calling the callback function for each type-definition, invoking
// the WIKIDB_GetSimilarity hook.  The callback function should return a number from
// 0 to 10, where 10 means that the data is DEFINITELY the type in question, and 0
// means it is definitely not.  The type that returns the highest number is the
// one that is chosen.  If there is more than one type with the same highest number,
// then the first one that was defined wins. The function will return the name of the
// matched type (as a string).
	function GuessType($FieldValue) {
		$Types = WikiDB_TypeHandler::pManageTypes(pWIKIDB_GetAllTypes);
		$ScoredTypes = array();

		foreach ($Types as $TypeName => $CallbackFunction) {
			if (is_callable($CallbackFunction)) {
				$ScoredTypes[$TypeName] = call_user_func($CallbackFunction,
														 WIKIDB_GetSimilarity,
														 $FieldValue,
														 array());
				if ($ScoredTypes[$TypeName] >= 10)
					return $TypeName;
			}
		}

		$LastScore = -1;
		$LastType = "";
		foreach ($ScoredTypes as $TypeName => $Score) {
			if ($Score > $LastScore) {
				$LastScore = $Score;
				$LastType = $TypeName;
			}
		}

		return $LastType;
	}

} // END: class WikiDB_TypeHandler
